#!/usr/bin/env python

import unittest
from pykeg.core import keg_sizes


class KegSizesTest(unittest.TestCase):
    def test_match(self):
        match = keg_sizes.find_closest_keg_size
        self.assertEqual('other', match(100.0))
        self.assertEqual('other', match(100000000.0))
        self.assertEqual('sixth', match(19570.6))
        self.assertEqual('sixth', match(19470.6))
        self.assertEqual('other', match(19460.6))

if __name__ == '__main__':
    unittest.main()



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Tasks for the Kegbot core."""

from pykeg.plugin import util as plugin_util
from pykeg import notification
from pykeg.core import stats
from pykeg.core import backup
from django.db import transaction

from pykeg.celery import app
from celery.utils.log import get_task_logger

logger = get_task_logger(__name__)


def schedule_tasks(events):
    """Synchronously schedules tasks related to the given events."""
    for plugin in plugin_util.get_plugins().values():
        try:
            plugin.handle_new_events(events)
        except Exception:
            logger.exception('Error dispatching events to plugin {}'.format(plugin.get_name()))
    notification.handle_new_system_events(events)


@app.task(name='build_stats', queue='stats', expires=60 * 60)
def build_stats(drink_id, rebuild_following):
    logger.info('build_stats drink_id={} rebuild_following={}'.format(
        drink_id, rebuild_following))
    with transaction.atomic():
        if rebuild_following:
            stats.rebuild_from_id(drink_id)
        else:
            stats.build_for_id(drink_id)


@app.task(name='build_backup', bind=True)
def build_backup(self):
    logger.info('build_backup')
    with transaction.atomic():
        backup.backup()



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for pykeg.web.api"""

from django.core import mail
from django.core.urlresolvers import reverse
from django.test import TransactionTestCase
from django.test.utils import override_settings
from pykeg.core import models
from pykeg.core import defaults
from pykeg.core.testutils import get_filename
from pykeg.core.util import get_version
from kegbot.util import kbjson

### Helper methods


def create_site():
    return defaults.set_defaults(set_is_setup=True, create_controller=True)


class BaseApiTestCase(TransactionTestCase):
    def get(self, subpath, data={}, follow=False, **extra):
        response = self.client.get('/api/%s' % subpath, data=data, follow=follow,
            **extra)
        return response, kbjson.loads(response.content)

    def post(self, subpath, data={}, follow=False, **extra):
        response = self.client.post('/api/%s' % subpath, data=data, follow=follow,
            **extra)
        return response, kbjson.loads(response.content)

    def delete(self, subpath, data={}, follow=False, **extra):
        response = self.client.delete('/api/%s' % subpath, data=data, follow=follow,
            **extra)
        return response, kbjson.loads(response.content)


class ApiClientNoSiteTestCase(BaseApiTestCase):
    def testNotSetUp(self):
        '''Api endpoints should all error out prior to site setup.'''

        endpoints = ('events/', 'taps/', 'kegs/', 'drinks/')
        for endpoint in endpoints:
            response, data = self.get(endpoint)
            self.assertEquals(data.meta.result, 'error')
            self.assertEquals(data.error.code, 'BadRequestError')

        create_site()

        # Ordinary results expected after site installed.
        for endpoint in endpoints:
            response, data = self.get(endpoint)
            self.assertEquals(data.meta.result, 'ok')


@override_settings(KEGBOT_BACKEND='pykeg.core.testutils.TestBackend')
class ApiClientTestCase(BaseApiTestCase):
    def setUp(self):
        self.site = create_site()
        self.admin = models.User.objects.create(username='admin', is_staff=True)
        self.admin.set_password('testpass')
        self.admin.save()

        self.normal_user = models.User.objects.create(username='normal_user', is_staff=True)
        self.normal_user.set_password('testpass')
        self.normal_user.save()

        self.apikey = models.ApiKey.objects.create(user=self.admin, key='123')
        self.bad_apikey = models.ApiKey.objects.create(user=self.normal_user, key='456')

    def test_defaults(self):
        empty_endpoints = ('events/', 'kegs/')
        for endpoint in empty_endpoints:
            response, data = self.get(endpoint)
            self.assertEquals(data.objects, [])

        response, data = self.get('taps/')
        taps = data.objects
        self.assertEquals(2, len(taps))
        self.assertEquals('Main Tap', taps[0].name)
        self.assertEquals('kegboard.flow0', taps[0].meter_name)
        self.assertEquals('Second Tap', taps[1].name)
        self.assertEquals('kegboard.flow1', taps[1].meter_name)

        for tap in taps:
            response1, data1 = self.get('taps/%s' % tap.meter_name)
            self.assertEquals(data1.meta.result, 'ok')

            response2, data2 = self.get('taps/%s' % tap.id)
            self.assertEquals(data2.meta.result, 'ok')

            self.assertEquals(data1, data2)

    def test_api_access(self):
        endpoint = 'users/'

        # No API key.
        response, data = self.get(endpoint)
        self.assertEquals(data.meta.result, 'error')
        self.assertEquals(data.error.code, 'NoAuthTokenError')

        # Non-existent key.
        response, data = self.get(endpoint, HTTP_X_KEGBOT_API_KEY='foobar')
        self.assertEquals(data.meta.result, 'error')
        self.assertEquals(data.error.code, 'BadApiKeyError')

        # Key exists, non-superuser.
        response, data = self.get(endpoint, HTTP_X_KEGBOT_API_KEY=self.bad_apikey.key)
        self.assertEquals(data.meta.result, 'ok')

        # Finally ok.
        response, data = self.get(endpoint, HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'ok')

        endpoint = 'events/'

        # Alter privacy and compare.
        response, data = self.get(endpoint)
        self.assertEquals(data.meta.result, 'ok')
        self.site.privacy = 'members'
        self.site.save()

        response, data = self.get(endpoint)
        self.assertEquals(data.meta.result, 'error')

        self.client.login(username='admin', password='testpass')
        response, data = self.get(endpoint)
        self.assertEquals(data.meta.result, 'ok')

        # Alert to staff-only.
        self.site.privacy = 'staff'
        self.site.save()
        response, data = self.get(endpoint)
        self.assertEquals(data.meta.result, 'ok')

        self.client.logout()
        response, data = self.get(endpoint)
        self.assertEquals(data.meta.result, 'error')
        self.assertEquals(data.error.code, 'NoAuthTokenError')

        # Login endpoint works despite privacy settings
        for ep in ('login', 'login/', 'v1/login', 'v1/login/'):
            response, data = self.post(ep, data={'username': 'admin', 'password': 'testpass'})
            self.assertEquals(data.meta.result, 'ok')

        for ep in ('version/', 'v1/version/'):
            response, data = self.get(ep)
            self.assertEquals(data.meta.result, 'ok')

    def test_record_drink(self):
        response, data = self.get('taps/1')
        self.assertEquals(data.meta.result, 'ok')
        self.assertEquals(data.object.get('current_keg'), None)

        response, data = self.get('drinks/last')
        self.assertEquals(response.status_code, 404)
        self.assertEquals(data.meta.result, 'error')

        new_keg_data = {
            'keg_size': 'half-barrel',
            'beverage_name': 'Test Brew',
            'producer_name': 'Test Producer',
            'style_name': 'Test Style,'
        }
        response, data = self.post('taps/1/activate', data=new_keg_data)
        self.assertEquals(data.meta.result, 'error')
        self.assertEquals(data.error.code, 'NoAuthTokenError')

        response, data = self.post('taps/1/activate', data=new_keg_data,
            HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'ok')
        self.assertIsNotNone(data.object.get('current_keg'))

        response, data = self.post('taps/1', data={'ticks': 1000})
        self.assertEquals(data.meta.result, 'error')
        self.assertEquals(data.error.code, 'NoAuthTokenError')

        response, data = self.post('taps/1', HTTP_X_KEGBOT_API_KEY=self.apikey.key,
            data={'ticks': 1000, 'username': self.normal_user.username})
        self.assertEquals(data.meta.result, 'ok')
        drink = data.object

        response, data = self.get('drinks/last')
        self.assertEquals(response.status_code, 200)
        self.assertEquals(data.meta.result, 'ok')
        self.assertEquals(data.object.id, drink.id)

        response, data = self.get('status', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'ok')
        users = data.object.get('active_users', [])
        self.assertEquals(1, len(users))
        active_user = users[0]
        self.assertEquals(self.normal_user.username, active_user.username)

    @override_settings(EMAIL_BACKEND='django.core.mail.backends.locmem.EmailBackend')
    @override_settings(EMAIL_FROM_ADDRESS='test-from@example')
    def test_registration(self):
        response, data = self.post('new-user/', data={'username': 'newuser', 'email': 'foo@example.com'})
        self.assertEquals(data.meta.result, 'error')
        self.assertEquals(data.error.code, 'NoAuthTokenError')

        self.assertEquals(0, len(mail.outbox))

        response, data = self.post('new-user/', data={'username': 'newuser', 'email': 'foo@example.com'},
            HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'ok')
        self.assertEquals(1, len(mail.outbox))

        msg = mail.outbox[0]
        self.assertEquals('[My Kegbot] Complete your registration', msg.subject)
        self.assertEquals(['foo@example.com'], msg.to)
        self.assertEquals('test-from@example', msg.from_email)

        # Simulate clicking on the activation link.
        user = models.User.objects.get(username='newuser')
        self.assertIsNotNone(user.activation_key)
        activation_url = reverse('activate-account', args=(),
            kwargs={'activation_key': user.activation_key})
        self.client.logout()
        response = self.client.get(activation_url)
        self.assertContains(response, 'Choose a Password', status_code=200)

    def test_pictures(self):
        image_data = open(get_filename('test_image_800x600.png'))
        response, data = self.post('pictures/', data={'photo': image_data}, HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'ok')

        picture = data['object']
        picture_url = picture['url']
        self.assertTrue(picture_url.startswith('http://localhost:1234/media/'))

    def test_controller_data(self):
        for endpoint in ('controllers', 'flow-meters'):
            response, data = self.get(endpoint)
            self.assertEquals(data.meta.result, 'error')
            self.assertEquals(data.error.code, 'NoAuthTokenError')

        response, data = self.get('controllers', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'ok')
        expected = {
            'objects': [
                {
                    'id': 1,
                    'name': 'kegboard',
                }
            ],
            'meta': {
                'result': 'ok',
            }
        }
        self.assertEquals(expected, data)

        response, data = self.get('flow-meters', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'ok')
        expected = {
            'objects': [
                {
                    'id': 1,
                    'ticks_per_ml': 2.724,
                    'port_name': 'flow0',
                    'controller': {
                        'id': 1,
                        'name': 'kegboard',
                    },
                    'name': 'kegboard.flow0'
                },
                {
                    'id': 2,
                    'ticks_per_ml': 2.724,
                    'port_name': 'flow1',
                    'controller': {
                        'id': 1,
                        'name': 'kegboard'
                    },
                    'name': 'kegboard.flow1'
                },
            ],
            'meta': {'result': 'ok'}
        }
        self.assertEquals(expected, data)

        response, data = self.get('flow-toggles', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'ok')
        expected = {
            'objects': [
                {
                    'id': 1,
                    'port_name': 'relay0',
                    'controller': {
                        'id': 1,
                        'name': 'kegboard',
                    },
                    'name': 'kegboard.relay0'
                },
                {
                    'id': 2,
                    'port_name': 'relay1',
                    'controller': {
                        'id': 1,
                        'name': 'kegboard'
                    },
                    'name': 'kegboard.relay1'
                },
            ],
            'meta': {'result': 'ok'}
        }
        self.assertEquals(expected, data)

    def test_add_remove_meters(self):
        response, data = self.get('taps/1', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'ok')
        self.assertEquals(data.object.meter.id, 1)
        original_data = data

        response, data = self.post('taps/1/disconnect-meter', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'ok')
        self.assertEquals(data.object.get('meter'), None)

        response, data = self.post('taps/1/connect-meter', HTTP_X_KEGBOT_API_KEY=self.apikey.key,
            data={'meter': 1})
        self.assertEquals(data.meta.result, 'ok')
        self.assertEquals(data.object.meter.id, 1)
        self.assertEquals(original_data, data)

    def test_add_remove_toggles(self):
        response, data = self.get('taps/1', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'ok')
        self.assertEquals(data.object.toggle.id, 1)

        response, data = self.post('taps/1/disconnect-toggle', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'ok')
        self.assertEquals(data.object.get('toggle'), None)

        response, data = self.post('taps/1/connect-toggle', HTTP_X_KEGBOT_API_KEY=self.apikey.key,
            data={'toggle': 1})
        response, data = self.get('taps/1', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'ok')
        self.assertIsNotNone(data.object.get('toggle'))
        self.assertEquals(data.object.toggle.id, 1)

    def test_get_version(self):
        response, data = self.get('version')
        self.assertEquals(data.meta.result, 'ok')
        self.assertEquals(data.object.get('server_version'), get_version())

    def test_devices(self):
        ### Perform a device link.
        response, data = self.post('devices/link', data={'name': 'Test Device'})
        self.assertEquals(data.meta.result, 'ok')
        code = data.object.code

        response, data = self.get('devices/link/status/' + code)
        self.assertEquals(data.meta.result, 'ok')
        self.assertEquals(False, data.object.linked)

        self.client.login(username='admin', password='testpass')
        response = self.client.post('/kegadmin/devices/link/', data={'code': code}, follow=True)
        self.client.logout()
        self.assertEquals(response.status_code, 200)

        response, data = self.get('devices/link/status/' + code)
        self.assertEquals(data.meta.result, 'ok')
        self.assertEquals(True, data.object.linked)
        api_key = data.object.get('api_key')
        self.assertIsNotNone(api_key)

        key_obj = models.ApiKey.objects.get(key=api_key)
        self.assertIsNotNone(key_obj.device)
        self.assertEquals('Test Device', key_obj.device.name)

        ### Confirm device key is gone.
        response, data = self.get('devices/link/status/' + code)
        self.assertEquals(response.status_code, 404)

    ### Kegbot object tests

    def test_auth_tokens(self):
        response, data = self.get('auth-tokens/nfc/deadbeef', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(data.meta.result, 'error')
        self.assertEquals(response.status_code, 404)

        response, data = self.post('auth-tokens/nfc/deadbeef/assign', HTTP_X_KEGBOT_API_KEY=self.apikey.key,
            data={'username': self.normal_user.username})
        self.assertEquals(data.meta.result, 'ok')
        self.assertEquals(data.object.auth_device, 'nfc')
        self.assertEquals(data.object.token_value, 'deadbeef')
        self.assertEquals(data.object.username, 'normal_user')
        self.assertEquals(response.status_code, 200)

    def test_controllers(self):
        """List, create, update, and delete controllers."""
        # List controllers.
        response, data = self.get('controllers', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 200)
        self.assertEquals(1, len(data.objects))
        self.assertEquals('kegboard', data.objects[0]['name'])

        # Create a new controller.
        response, data = self.post('controllers', HTTP_X_KEGBOT_API_KEY=self.apikey.key,
            data={
                'name': 'Test Controller',
                'model_name': 'Test Model',
                'serial_number': 'Test Serial'
            })
        self.assertEquals(response.status_code, 200)
        self.assertEquals('Test Controller', data.object.name)
        self.assertEquals('Test Model', data.object.model_name)
        self.assertEquals('Test Serial', data.object.serial_number)

        # Fetch controller.
        new_controller_id = data.object.id
        response, data = self.get('controllers/' + str(new_controller_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 200)
        self.assertEquals('Test Controller', data.object.name)
        self.assertEquals('Test Model', data.object.model_name)
        self.assertEquals('Test Serial', data.object.serial_number)

        # Update controller
        response, data = self.post('controllers/' + str(new_controller_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key,
            data={
                'name': 'Test Controller+',
                'model_name': 'Test Model+',
                'serial_number': 'Test Serial+'})
        self.assertEquals(response.status_code, 200)
        self.assertEquals('Test Controller+', data.object.name)
        self.assertEquals('Test Model+', data.object.model_name)
        self.assertEquals('Test Serial+', data.object.serial_number)

        response, data = self.get('controllers/' + str(new_controller_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 200)
        self.assertEquals('Test Controller+', data.object.name)
        self.assertEquals('Test Model+', data.object.model_name)
        self.assertEquals('Test Serial+', data.object.serial_number)

        # Delete controller
        response, data = self.delete('controllers/' + str(new_controller_id))
        self.assertEquals(response.status_code, 401)
        response, data = self.delete('controllers/' + str(new_controller_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 200)
        response, data = self.get('controllers/' + str(new_controller_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 404)

    def test_flow_meters(self):
        """List, create, update, and delete flow meters."""
        # List flow meters.
        response, data = self.get('flow-meters', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 200)
        self.assertEquals(2, len(data.objects))
        self.assertEquals('kegboard.flow0', data.objects[0]['name'])
        self.assertEquals('flow0', data.objects[0]['port_name'])
        self.assertEquals('kegboard.flow1', data.objects[1]['name'])
        self.assertEquals('flow1', data.objects[1]['port_name'])

        # Create a new meter.
        controller = models.Controller.objects.all()[0]
        response, data = self.post('flow-meters', HTTP_X_KEGBOT_API_KEY=self.apikey.key,
            data={
                'port_name': 'flow-test',
                'ticks_per_ml': 3.45,
                'controller': controller.id,
            })
        self.assertEquals(response.status_code, 200)
        self.assertEquals('kegboard.flow-test', data.object.name)
        self.assertEquals('flow-test', data.object.port_name)
        self.assertEquals(3.45, data.object.ticks_per_ml)
        self.assertEquals(controller.name, data.object.controller.name)

        # Fetch meter.
        new_meter_id = data.object.id
        response, data = self.get('flow-meters/' + str(new_meter_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 200)
        self.assertEquals('kegboard.flow-test', data.object.name)
        self.assertEquals('flow-test', data.object.port_name)
        self.assertEquals(3.45, data.object.ticks_per_ml)
        self.assertEquals(controller.name, data.object.controller.name)

        # Update meter
        response, data = self.post('flow-meters/' + str(new_meter_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key,
        data={
            'ticks_per_ml': 5.67,
        })
        self.assertEquals(response.status_code, 200)
        self.assertEquals('kegboard.flow-test', data.object.name)
        self.assertEquals(5.67, data.object.ticks_per_ml)
        self.assertEquals(controller.name, data.object.controller.name)

        # Delete meter
        response, data = self.delete('flow-meters/' + str(new_meter_id))
        self.assertEquals(response.status_code, 401)
        response, data = self.delete('flow-meters/' + str(new_meter_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 200)
        response, data = self.get('flow-meters/' + str(new_meter_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 404)

    def test_flow_toggles(self):
        """List, create, and delete flow toggles."""
        # List flow toggles.
        response, data = self.get('flow-toggles', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 200)
        self.assertEquals(2, len(data.objects))
        self.assertEquals('kegboard.relay0', data.objects[0]['name'])
        self.assertEquals('relay0', data.objects[0]['port_name'])
        self.assertEquals('kegboard.relay1', data.objects[1]['name'])
        self.assertEquals('relay1', data.objects[1]['port_name'])

        # Create a new toggle.
        controller = models.Controller.objects.all()[0]
        response, data = self.post('flow-toggles', HTTP_X_KEGBOT_API_KEY=self.apikey.key,
            data={
                'port_name': 'toggle-test',
                'controller': controller.id,
            })
        self.assertEquals(response.status_code, 200)
        self.assertEquals('kegboard.toggle-test', data.object.name)
        self.assertEquals('toggle-test', data.object.port_name)
        self.assertEquals(controller.name, data.object.controller.name)

        # Fetch toggle.
        new_toggle_id = data.object.id
        response, data = self.get('flow-toggles/' + str(new_toggle_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 200)
        self.assertEquals('kegboard.toggle-test', data.object.name)
        self.assertEquals('toggle-test', data.object.port_name)
        self.assertEquals(controller.name, data.object.controller.name)

        # Delete toggle.
        response, data = self.delete('flow-toggles/' + str(new_toggle_id))
        self.assertEquals(response.status_code, 401)
        response, data = self.delete('flow-toggles/' + str(new_toggle_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 200)
        response, data = self.get('flow-toggles/' + str(new_toggle_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 404)

    def test_taps(self):
        """List, create, and delete taps."""
        # List flow toggles.
        response, data = self.get('taps', HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 200)
        self.assertEquals(2, len(data.objects))
        self.assertEquals('Main Tap', data.objects[0]['name'])
        self.assertEquals('Second Tap', data.objects[1]['name'])

        # Create a new toggle.
        response, data = self.post('taps', HTTP_X_KEGBOT_API_KEY=self.apikey.key,
            data={
                'name': 'Test Tap',
            })
        self.assertEquals(response.status_code, 200)
        self.assertEquals('Test Tap', data.object.name)

        # Fetch tap.
        tap_id = data.object.id
        response, data = self.get('taps/' + str(tap_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 200)
        self.assertEquals('Test Tap', data.object.name)

        # Delete tap.
        response, data = self.delete('taps/' + str(tap_id))
        self.assertEquals(response.status_code, 401)
        response, data = self.delete('taps/' + str(tap_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 200)
        response, data = self.get('taps/' + str(tap_id), HTTP_X_KEGBOT_API_KEY=self.apikey.key)
        self.assertEquals(response.status_code, 404)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from . import models

METER_NAME_0 = 'kegboard.flow0'
METER_NAME_1 = 'kegboard.flow1'


def db_is_installed():
    return models.KegbotSite.objects.all().count() > 0


class AlreadyInstalledError(Exception):
    """Thrown when database is already installed."""


def set_defaults(force=False, set_is_setup=False, create_controller=False):
    """Creates a new site and sets defaults, returning that site."""
    if not force and db_is_installed():
        raise AlreadyInstalledError("Database is already installed.")

    site = models.KegbotSite.get()
    if set_is_setup and not site.is_setup:
        site.is_setup = True
        site.save()

    models.User.objects.create_user('guest')

    tap_0 = models.KegTap.objects.create(name='Main Tap')

    # Prior to 0.9.23, controller objects were created automatically.
    # This behavior is available for unittests which request it.
    if create_controller:
        tap_1 = models.KegTap.objects.create(name='Second Tap')
        controller = models.Controller.objects.create(name='kegboard')

        models.FlowMeter.objects.create(controller=controller, port_name='flow0',
            tap=tap_0)
        models.FlowMeter.objects.create(controller=controller, port_name='flow1',
            tap=tap_1)
        models.FlowToggle.objects.create(controller=controller, port_name='relay0',
            tap=tap_0)
        models.FlowToggle.objects.create(controller=controller, port_name='relay1',
            tap=tap_1)

    return site



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Protocol for associating new frontend devices with a Kegbot server.

The protocol consists of three phases, ultimately granting the requesting
device (client) a new API key to Kegbot Server.

Protocol:

    1. Client initiates a new pairing session by POSTing to
       `/api/devices/link`.

    2. Server generates a new opaque and unguessable short-lived
       pairing token (`start_link()`) and returns it to the client.

    3. Client displays the pairing token on its screen, and continuously
       polls its status via `/api/devices/status/<token>` (`get_status()`).

    4. User logs in to the web interface and confirms the request by
       entering the same access token into a form, causing the token
       to be marked as validated (`confirm_link()`).

    5. On next subsequent poll of `/api/devices/status/<token>`, the
       request handler discovers the token has been confirm, creates
       a new Device record and ApiKey, and returns the ApiKey.
"""

import random
from django.core.cache import cache
from pykeg.core import models

CACHE_PREFIX = 'devicelink:'
CACHE_TIMEOUT = 60 * 60

# Letters used to build a new device link code.  Currently
# A-Z,0-9 sans visually-ambiguous characters {I, 1, O, 0}.
CODE_LETTERS = 'ABCDEFGHJKLMNPQRSTUVWXYZ23456789'

# Default code size generated by _build_code.  Entropy of the pairing
# code is:
#
#   len(CODE_LETTERS) ** DEFAULT_CODE_SIZE
#
# Entropy is only needed to prevent collisions of active pairing
# sessions, and not for security of the protocol (since an attacker
# still needs a logged-in admin session cookie to accept a
# guessed code).
DEFAULT_CODE_SIZE = 6


class LinkExpiredException(Exception):
    """Link code not found."""


def _build_code(size=DEFAULT_CODE_SIZE):
    code = ''.join(random.choice(CODE_LETTERS) for i in range(size))
    code = '{}-{}'.format(code[:(size / 2)], code[(size / 2):])
    return code


def _get_status(code):
    cache_key = ''.join((CACHE_PREFIX, code))
    status = cache.get(cache_key)
    return status


def _set_status(code, status):
    cache_key = ''.join((CACHE_PREFIX, code))
    cache.set(cache_key, status, timeout=CACHE_TIMEOUT)


def start_link(name):
    """Starts a new link, by allocating and returning a link code."""
    if not name:
        raise ValueError('Must specify a name.')
    code = _build_code()
    _set_status(code, {'name': name, 'linked': False})
    return code


def confirm_link(code):
    """Confirms a link, ie, flips its status to True."""
    status = _get_status(code)
    if not status:
        raise LinkExpiredException('Code not found')
    status['linked'] = True
    _set_status(code, status)
    return status


def get_status(code):
    """Gets status, finishing the link if true.

    Returns: string api key on link, None otherwise
    """
    status = _get_status(code)
    if not status:
        raise LinkExpiredException('Code not found')
    if status.get('linked'):
        cache_key = ''.join((CACHE_PREFIX, code))
        cache.delete(cache_key)
        device = models.Device.objects.create(name=status.get('name'))
        apikey = models.ApiKey.objects.create(description='Linked device.', device=device)
        return apikey.key
    return None



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.core.management.base import NoArgsCommand
from django.db import transaction

from pykeg.core import models
from pykeg.core import stats
from pykeg.core.management.commands.common import progbar


class Command(NoArgsCommand):
    help = u'Regenerate all cached stats.'
    args = '<none>'

    @transaction.atomic
    def handle(self, **options):

        num_drinks = models.Drink.objects.all().count()
        self.pos = 0

        def cb(results, self=self):
            self.pos += 1
            progbar('regenerating stats', self.pos, num_drinks)

        stats.invalidate_all()
        stats.rebuild_from_id(0, cb=cb)

        print ''
        print 'done!'



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import json
import os.path
import random
import datetime
from django.conf import settings
from collections import deque
from uuid import uuid1

from django.core.files import File
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError
from django.db import transaction
from django.utils import timezone

from pykeg.contrib.demomode.views import random_item
from pykeg.backend import get_kegbot_backend
from pykeg.core import defaults
from pykeg.core import models


NUM_SESSIONS = 30
MINUTES_IN_DAY = 60 * 24

MINIMUM_POUR_SIZE_ML = 100
MAXIMUM_POUR_SIZE_ML = 450

# Predictably generate same demo data.
RANDOM_SEED = 42
START_DATE = datetime.datetime(2014, 10, 1)

"""Generates a bunch of fake drinks from a configuration."""


class LoadDemoDataCommand(BaseCommand):
    args = '<demo_data_dir>'
    help = 'Inserts demo data into the database'

    def handle(self, *args, **options):
        if len(args) < 1:
            raise CommandError('Must specify demo data directory')

        data_dir = os.path.abspath(args[0])
        demo_data = self.load_demo_data(data_dir)

        settings.CELERY_ALWAYS_EAGER = True

        with transaction.atomic():
            if models.User.objects.all().count() or models.Drink.objects.all().count():
                raise CommandError('Database is already loaded.')

            defaults.set_defaults(set_is_setup=True, create_controller=True)

            # Install demo data.
            for username in demo_data['drinkers']:
                user = models.User.objects.create_superuser(username=username, email=None, password=username)

                picture_path = os.path.join(data_dir, 'pictures', 'drinkers', username, 'mugshot.png')
                if os.path.exists(picture_path):
                    p = self.create_picture(picture_path, user=user)
                    user.mugshot = p
                    user.save()

            for beverage in demo_data['beverages']:
                brewer, _ = models.BeverageProducer.objects.get_or_create(name=beverage['brewer_name'])
                beer, _ = models.Beverage.objects.get_or_create(name=beverage['name'], beverage_type='beer',
                    producer=brewer, style=beverage['style'])

                picture_path = beverage.get('image')
                if picture_path:
                    picture = self.create_picture(os.path.join(data_dir, picture_path))
                    beer.picture = picture
                    beer.save()
                else:
                    pass

            random.seed(RANDOM_SEED)
            sessions = self.build_random_sessions(models.User.objects.all(),
                NUM_SESSIONS, demo_data['shouts'])

            minutes = sessions[-1][0]
            minutes += MINUTES_IN_DAY
            minutes -= (minutes % (MINUTES_IN_DAY))

            session_start = timezone.make_aware(START_DATE, timezone.utc)
            session_start -= datetime.timedelta(hours=session_start.hour + 4,
                minutes=session_start.minute,
                seconds=session_start.second)
            session_start -= datetime.timedelta(minutes=minutes)

            for minute, drinker, volume_ml, shout in sessions:
                date = session_start + datetime.timedelta(minutes=minute)

                picture_path = None
                if random.random() >= 0.25:
                    d = demo_data['pictures'].get(drinker.username)
                    if d:
                        picture_path = d[0]
                        d.rotate()
                self.do_pour(drinker, date, volume_ml, shout, picture_path)

            print 'Demo data loaded.'

    def do_pour(self, user, when, volume_ml, shout, picture_path):
        be = get_kegbot_backend()
        tap = random_item(models.KegTap)

        # End keg if it's near empty.
        if tap.current_keg and tap.current_keg.remaining_volume_ml() < volume_ml:
            be.end_keg(tap)

        # Start keg if the tap is idle.
        if not tap.current_keg:
            beverage = random_item(models.Beverage)
            be.start_keg(tap, beverage=beverage)

        drink = be.record_drink(tap, ticks=0, volume_ml=volume_ml,
            username=user.username, pour_time=when, shout=shout)

        if picture_path:
            p = self.create_picture(picture_path, user=user,
                session=drink.session, keg=drink.keg)
            drink.picture = p
            drink.save()
        return drink

    def create_picture(self, path, **kwargs):
        f = File(open(path, 'r'))
        p = models.Picture(image=f, **kwargs)
        name = '%s%s' % (uuid1(), os.path.splitext(f.name)[1])
        p.image.save(name, f)
        p.save()
        return p

    def build_random_sessions(self, all_drinkers, count, shouts):
        MIN_POUR_ML = 200
        MAX_POUR_ML = 500

        pours = []
        minute = 0

        for session_number in xrange(count):
            session = []
            num_drinkers = random.randint(1, len(all_drinkers))
            drinkers = random.sample(all_drinkers, num_drinkers)

            # Drink a little more when others are joining
            session_max = 1000 + min(2000, 500 * len(drinkers))

            for drinker in drinkers:
                subminute = random.randint(0, 15)
                session_total = random.randint(500, session_max)
                session_remain = session_total

                while session_remain > 0:
                    if session_remain <= MIN_POUR_ML:
                        drink_volume = session_remain
                    else:
                        drink_volume = random.randint(MIN_POUR_ML, MAX_POUR_ML)

                    shout = ''
                    if shouts and random.random() >= 0.60:
                        shout = shouts[0]
                        shouts.rotate(1)

                    session.append((minute + subminute, drinker, drink_volume, shout))
                    subminute += int(drink_volume / 40) + random.randint(0, 30)  # 40 mL/minute min plus noise
                    session_remain -= drink_volume

            # session.sort()
            # for minute, drinker, volume, shout in session:
            #     print '%08s %016s %04s  %s' % (minute, drinker, volume, shout)
            # print ''

            pours += session

            # wait 1 - 3 days for next session
            minute += random.randint(1, 3) * MINUTES_IN_DAY
            minute -= (minute % (MINUTES_IN_DAY))
            minute += random.randint(-120, 120)

        pours.sort()
        return pours

    def load_demo_data(self, path):
        demo_data_file = os.path.join(path, 'demo_data.json')
        if not os.path.isfile(demo_data_file):
            raise CommandError('Missing demo_data file: %s' % demo_data_file)

        demo_data = json.loads(open(demo_data_file).read())
        demo_data['pictures'] = {}

        for drinker_name in demo_data['drinkers']:
            demo_data['pictures'][drinker_name] = deque()

            pictures_dir = os.path.join(path, 'pictures', 'drinkers', drinker_name)
            if not os.path.isdir(pictures_dir):
                continue

            for picture_name in (p for p in os.listdir(pictures_dir) if not p.startswith('.')):
                if picture_name == 'mugshot.png':
                    continue
                picture_path = os.path.join(pictures_dir, picture_name)
                demo_data['pictures'][drinker_name].append(picture_path)
            random.shuffle(demo_data['pictures'][drinker_name])

        demo_data['shouts'] = deque(demo_data['shouts'])
        random.shuffle(demo_data['shouts'])

        return demo_data

Command = LoadDemoDataCommand



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for pykeg.web.api.devicelink"""

from django.core.cache import cache
from django.test import TransactionTestCase
from pykeg.core import models
from pykeg.web.api import devicelink


class DevicelinkTest(TransactionTestCase):
    def test_pairing(self):
        code = devicelink.start_link('test1')
        key = ''.join((devicelink.CACHE_PREFIX, code))
        saved = cache.get(key)
        self.assertEqual({'name': 'test1', 'linked': False}, saved)

        status = devicelink.get_status(code)
        self.assertIsNone(status)

        devicelink.confirm_link(code)
        saved = cache.get(key)
        self.assertEqual({'name': 'test1', 'linked': True}, saved)

        status = devicelink.get_status(code)
        self.assertIsNotNone(status)
        apikey = models.ApiKey.objects.get(key=status)
        self.assertEqual('test1', apikey.device.name)

        # Entry has been deleted.
        self.assertRaises(devicelink.LinkExpiredException, devicelink.get_status,
            code)

        self.assertRaises(devicelink.LinkExpiredException, devicelink.get_status,
            'bogus-code')

    def test_build_code(self):
        code = devicelink._build_code(6)
        self.assertEqual(7, len(code))
        self.assertEqual('-', code[3])

        code = devicelink._build_code(3)
        self.assertEqual(4, len(code))
        self.assertEqual('-', code[1])



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""General tests for the web interface."""

from django.core import mail
from django.test import TransactionTestCase
from django.test.utils import override_settings
from django.core.urlresolvers import reverse

from pykeg.backend import get_kegbot_backend
from pykeg.core import models
from pykeg.core import defaults


@override_settings(KEGBOT_BACKEND='pykeg.core.testutils.TestBackend')
class KegwebTestCase(TransactionTestCase):
    def setUp(self):
        self.client.logout()
        defaults.set_defaults(set_is_setup=True, create_controller=True)

    def testBasicEndpoints(self):
        for endpoint in ('/kegs/', '/stats/', '/drinkers/guest/', '/drinkers/guest/sessions/'):
            response = self.client.get(endpoint)
            self.assertEquals(200, response.status_code)

        for endpoint in ('/sessions/',):
            response = self.client.get(endpoint)
            self.assertEquals(404, response.status_code)

        b = get_kegbot_backend()
        keg = b.start_keg('kegboard.flow0', beverage_name='Unknown', producer_name='Unknown',
            beverage_type='beer', style_name='Unknown')
        self.assertIsNotNone(keg)
        response = self.client.get('/kegs/')
        self.assertEquals(200, response.status_code)

        d = b.record_drink('kegboard.flow0', ticks=100)
        drink_id = d.id

        response = self.client.get('/d/%s' % drink_id, follow=True)
        self.assertRedirects(response, '/drinks/%s' % drink_id, status_code=301)

        session_id = d.session.id
        response = self.client.get('/s/%s' % session_id, follow=True)
        self.assertRedirects(response, d.session.get_absolute_url(), status_code=301)

    def testShout(self):
        b = get_kegbot_backend()
        b.start_keg('kegboard.flow0', beverage_name='Unknown', producer_name='Unknown',
            beverage_type='beer', style_name='Unknown')
        d = b.record_drink('kegboard.flow0', ticks=123, shout='_UNITTEST_')
        response = self.client.get(d.get_absolute_url())
        self.assertContains(response, '<p>_UNITTEST_</p>', status_code=200)

    def test_privacy(self):
        b = get_kegbot_backend()
        keg = b.start_keg('kegboard.flow0', beverage_name='Unknown', producer_name='Unknown',
            beverage_type='beer', style_name='Unknown')
        self.assertIsNotNone(keg)
        d = b.record_drink('kegboard.flow0', ticks=100)

        # URLs to expected contents
        urls = {
            '/kegs/': 'Keg List',
            '/stats/': 'System Stats',
            '/sessions/': 'All Sessions',
            '/kegs/{}'.format(keg.id): 'Keg {}'.format(keg.id),
            '/drinks/{}'.format(d.id): 'Drink {}'.format(d.id),
        }

        def test_urls(expect_fail, urls=urls):
            for url, expected_content in urls.iteritems():
                response = self.client.get(url)
                if expect_fail:
                    self.assertNotContains(response, expected_content, status_code=401,
                            msg_prefix=url)
                else:
                    self.assertContains(response, expected_content, status_code=200,
                            msg_prefix=url)

        b = get_kegbot_backend()
        user = b.create_new_user('testuser', 'test@example.com', password='1234')

        kbsite = models.KegbotSite.get()
        self.client.logout()

        # Public mode.
        test_urls(expect_fail=False)

        # Members-only.
        kbsite.privacy = 'members'
        kbsite.save()
        test_urls(expect_fail=True)
        logged_in = self.client.login(username='testuser', password='1234')
        self.assertTrue(logged_in)
        test_urls(expect_fail=False)

        # Staff-only
        kbsite.privacy = 'staff'
        kbsite.save()

        test_urls(expect_fail=True)
        user.is_staff = True
        user.save()
        test_urls(expect_fail=False)
        self.client.logout()
        test_urls(expect_fail=True)

    def test_whitelisted_urls(self):
        """Verify always-accessible URLs."""
        urls = (
            '/accounts/password/reset/',
            '/accounts/register/',
            '/accounts/login/',
        )

        for url in urls:
            response = self.client.get(url)
            self.assertNotContains(response, 'denied', status_code=200,
                    msg_prefix=url)

    def test_activation(self):
        b = get_kegbot_backend()
        kbsite = models.KegbotSite.get()
        self.assertEqual('public', kbsite.privacy)

        user = b.create_new_user('testuser', 'test@example.com')
        self.assertIsNotNone(user.activation_key)
        self.assertFalse(user.has_usable_password())

        activation_key = user.activation_key
        self.assertIsNotNone(activation_key)

        activation_url = reverse('activate-account', args=(),
            kwargs={'activation_key': activation_key})

        # Activation works regardless of privacy settings.
        self.client.logout()
        response = self.client.get(activation_url)
        self.assertContains(response, 'Choose a Password', status_code=200)

        kbsite.privacy = 'staff'
        kbsite.save()
        response = self.client.get(activation_url)
        self.assertContains(response, 'Choose a Password', status_code=200)

        kbsite.privacy = 'members'
        kbsite.save()
        response = self.client.get(activation_url)
        self.assertContains(response, 'Choose a Password', status_code=200)

        # Activate the account.
        form_data = {
            'password': '123',
            'password2': '123',
        }

        response = self.client.post(activation_url, data=form_data, follow=True)
        self.assertContains(response, 'Your account has been activated!', status_code=200)
        user = models.User.objects.get(pk=user.id)
        self.assertIsNone(user.activation_key)

    @override_settings(EMAIL_BACKEND='django.core.mail.backends.locmem.EmailBackend')
    @override_settings(EMAIL_FROM_ADDRESS='test-from@example')
    def test_registration(self):
        kbsite = models.KegbotSite.get()
        self.assertEqual('public', kbsite.privacy)
        self.assertEqual('public', kbsite.registration_mode)

        response = self.client.get('/accounts/register/')
        self.assertContains(response, 'Register New Account', status_code=200)

        response = self.client.post('/accounts/register/',
            data={
                'username': 'newuser',
                'password1': '1234',
                'password2': '1234',
                'email': 'test2@example.com',
            }, follow=True)
        self.assertRedirects(response, '/account/')
        self.assertContains(response, 'Hello, newuser')
        self.assertEqual(1, len(mail.outbox))

        msg = mail.outbox[0]
        self.assertEqual(['test2@example.com'], msg.to)
        self.assertTrue('To log in to your account, please click here' in msg.body)

        response = self.client.post('/accounts/register/',
            data={
                'username': 'newuser',
                'password1': '1234',
                'password2': '1234',
                'email': 'test2@example.com',
            }, follow=False)
        self.assertContains(response, 'User with this Username already exists',
            status_code=200)

        response = self.client.post('/accounts/register/',
            data={
                'username': 'newuser 2',
                'password1': '1234',
                'password2': '1234',
                'email': 'test2@example.com',
            }, follow=False)
        self.assertContains(response, 'Enter a valid username',
            status_code=200)

        response = self.client.post('/accounts/register/',
            data={
                'username': 'newuser2',
                'password1': '1234',
                'password2': '1235',
                'email': 'test2@example.com',
            }, follow=False)
        print response
        self.assertContains(response, "The two password fields didn&#39;t match.",
            status_code=200)

    @override_settings(EMAIL_BACKEND='django.core.mail.backends.locmem.EmailBackend')
    @override_settings(EMAIL_FROM_ADDRESS='test-from@example')
    def test_registration_with_invite(self):
        kbsite = models.KegbotSite.get()
        kbsite.registration_mode = 'staff-invite-online'
        kbsite.save()

        response = self.client.get('/accounts/register/')
        self.assertContains(response, 'Invitation Required', status_code=401)

        response = self.client.get('/accounts/register/?invite_code=1234')
        self.assertContains(response, 'Invitation Expired', status_code=401)

        models.Invitation.objects.create(invite_code='test', for_email='test@example.com')
        self.assertEqual(1, models.Invitation.objects.all().count())

        response = self.client.get('/accounts/register/?invite_code=test')
        self.assertContains(response, 'Register New Account', status_code=200)
        response = self.client.post('/accounts/register/',
            data={
                'username': 'newuser2',
                'password1': '1234',
                'password2': '1234',
                'email': 'test2@example.com',
            }, follow=True)
        self.assertRedirects(response, '/account/')
        self.assertContains(response, 'Hello, newuser2')
        self.assertEqual(0, models.Invitation.objects.all().count())

        response = self.client.get('/accounts/register/?invite_code=test')
        self.assertContains(response, 'Invitation Expired', status_code=401)

    def test_upgrade_bouncer(self):
        kbsite = models.KegbotSite.get()
        response = self.client.get('/')
        self.assertContains(response, 'My Kegbot', status_code=200)

        old_version = kbsite.server_version
        kbsite.server_version = '0.0.1'
        kbsite.save()
        response = self.client.get('/')
        self.assertContains(response, 'Upgrade Required', status_code=403)

        kbsite.server_version = old_version
        kbsite.is_setup = False
        kbsite.save()
        response = self.client.get('/')
        self.assertContains(response, 'Kegbot Offline', status_code=403)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import random

from django.contrib import messages
from django.db import transaction
from django.shortcuts import redirect
from django.utils.safestring import SafeString
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods

from pykeg.backend import get_kegbot_backend
from pykeg.core import models
from pykeg.contrib.demomode.util import random_item

MIN_KEG_VOLUME_ML = 800
RANDOM_POUR_RANGE_ML = (100, 500)


@csrf_exempt
@never_cache
@require_http_methods(["POST"])
@transaction.atomic
def summon_drinker(request):
    user = random_item(models.User)
    tap = random_item(models.KegTap)
    beverage = random_item(models.Beverage)

    if not tap or not user or not beverage:
        messages.error(request, 'No drinkers are available!')
        return redirect('kb-home')

    be = get_kegbot_backend()

    # End keg if it's near empty.
    if tap.current_keg and tap.current_keg.remaining_volume_ml() < MIN_KEG_VOLUME_ML:
        be.end_keg(tap)

    # Start keg if the tap is idle.
    if not tap.current_keg:
        be.start_keg(tap, beverage=beverage)

    volume_ml = random.randint(*RANDOM_POUR_RANGE_ML)

    drink = be.record_drink(tap, ticks=0, volume_ml=volume_ml,
        username=user.username)

    pictures = list(models.Picture.objects.filter(user=user))
    picture = random.choice(pictures) if pictures else None
    if picture:
        new_picture = models.Picture.objects.create(image=picture.image,
            user=user)
        drink.picture = new_picture
        drink.save()

    message = 'We poked <strong>%s</strong>, who just poured <a href="%s">Drink %s</a>.' % (
        user.username, drink.get_absolute_url(), drink.id)
    messages.success(request, SafeString(message))

    return redirect('kb-home')



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for backends module."""

from django.test import TestCase
from pykeg.backend import get_kegbot_backend
from pykeg.core import models
from pykeg.core import defaults
from django.test.utils import override_settings
from django.core import management

METER_NAME = 'kegboard.flow0'
FAKE_BEER_NAME = 'Testy Beer'
FAKE_BREWER_NAME = 'Unittest Brewery'
FAKE_BEER_STYLE = 'Test-Driven Pale Ale'

### Helper methods


@override_settings(KEGBOT_BACKEND='pykeg.core.testutils.TestBackend')
class BackendsFixtureTestCase(TestCase):
    """Test backened using fixture (demo) data."""

    @classmethod
    def setupClass(cls):
        management.call_command('loaddata', 'testdata/full_demo_site.json', verbosity=0)

    @classmethod
    def teardownClass(cls):
        management.call_command('flush', verbosity=0, interactive=False)

    def setUp(self):
        self.backend = get_kegbot_backend()

    def test_delete_keg(self):
        site = models.KegbotSite.get()
        original_stats = site.get_stats()
        self.assertEquals(667, original_stats['total_pours'])

        keg = models.Keg.objects.get(pk=2)
        keg_stats = keg.get_stats()
        keg_drinks = keg.drinks.all()
        self.assertEquals(188, len(keg_drinks))
        self.backend.cancel_keg(keg)

        stats = site.get_stats()
        self.assertEquals(667 - 188, stats['total_pours'])
        self.assertEquals(original_stats['total_volume_ml'] - keg_stats['total_volume_ml'],
            stats['total_volume_ml'])


@override_settings(KEGBOT_BACKEND='pykeg.core.testutils.TestBackend')
class BackendsTestCase(TestCase):
    """Test backend without canned data (legacy tests)."""

    def setUp(self):
        self.backend = get_kegbot_backend()
        defaults.set_defaults(set_is_setup=True, create_controller=True)

    def test_drink_management(self):
        """Test adding drinks."""
        tap = models.KegTap.get_from_meter_name(METER_NAME)
        tap.save()
        meter = tap.meter
        meter.ticks_per_ml = 2.2
        meter.save()
        keg = self.backend.start_keg(METER_NAME, beverage_name=FAKE_BEER_NAME,
                beverage_type='beer', producer_name=FAKE_BREWER_NAME,
                style_name=FAKE_BEER_STYLE)
        self.assertIsNotNone(keg)

        self.assertEquals(0, keg.served_volume())
        drink = self.backend.record_drink(METER_NAME, ticks=2200)
        self.assertIsNotNone(drink)
        self.assertEquals(2200, drink.ticks)
        self.assertAlmostEqual(1000.0, drink.volume_ml, places=3)
        keg = models.Keg.objects.get(pk=keg.id)
        self.assertAlmostEqual(1000.0, keg.served_volume(), places=3)
        self.assertEquals('', drink.shout)
        self.assertEquals(0, drink.duration)

        drink = self.backend.record_drink(METER_NAME, ticks=1100)
        keg = models.Keg.objects.get(pk=keg.id)
        self.assertIsNotNone(drink)
        self.assertEquals(1100, drink.ticks)
        self.assertAlmostEqual(500.0, drink.volume_ml, places=3)
        self.assertAlmostEqual(1500.0, keg.served_volume(), places=3)

        # Add with volume, taking precedence over ticks.
        drink = self.backend.record_drink(METER_NAME, ticks=1100, volume_ml=1)
        keg = models.Keg.objects.get(pk=keg.id)
        self.assertIsNotNone(drink)
        self.assertEquals(1100, drink.ticks)
        self.assertAlmostEqual(1.0, drink.volume_ml, places=3)
        self.assertAlmostEqual(1501.0, keg.served_volume(), places=3)

        # Add with a user.
        user = models.User.objects.create(username='testy')
        drink = self.backend.record_drink(METER_NAME, ticks=2200, username=user.username)
        self.assertIsNotNone(drink)
        self.assertEquals(user, drink.user)

    def test_drink_cancel(self):
        """Tests cancelling drinks."""
        keg = self.backend.start_keg(METER_NAME, beverage_name=FAKE_BEER_NAME,
                beverage_type='beer', producer_name=FAKE_BREWER_NAME,
                style_name=FAKE_BEER_STYLE)
        self.assertIsNotNone(keg)
        self.assertEquals(0, keg.served_volume())

        for i in xrange(10):
            self.backend.record_drink(METER_NAME, ticks=1, volume_ml=100)

        drinks = list(models.Drink.objects.all().order_by('id'))
        keg = models.Keg.objects.get(pk=keg.id)
        self.assertEquals(10, len(drinks))
        self.assertAlmostEqual(1000.0, keg.served_volume(), places=3)

        cancel_drink = drinks[-1]
        session = cancel_drink.session
        self.assertAlmostEqual(session.get_stats().total_volume_ml, 1000.0, places=3)

        self.backend.cancel_drink(drinks[-1])
        drinks = list(models.Drink.objects.all().order_by('id'))
        keg = models.Keg.objects.get(pk=keg.id)
        self.assertEquals(9, len(drinks))
        self.assertAlmostEqual(900.0, keg.served_volume(), places=3)
        session = models.DrinkingSession.objects.get(id=session.id)
        self.assertAlmostEqual(session.get_stats().total_volume_ml, 900.0, places=3)

        keg = models.Keg.objects.get(pk=keg.id)
        self.assertEquals(0, keg.spilled_ml)

        self.backend.cancel_drink(drinks[-1], spilled=True)
        keg = models.Keg.objects.get(pk=keg.id)
        drinks = list(models.Drink.objects.all().order_by('id'))
        self.assertEquals(8, len(drinks))
        self.assertAlmostEqual(800.0, keg.served_volume(), places=3)
        self.assertAlmostEqual(100.0, keg.spilled_ml, places=3)

        # Cancel all drinks and confirm the session is deleted.
        num_sessions = models.DrinkingSession.objects.all().count()
        first_drink, other_drinks = drinks[0], drinks[1:]
        for d in other_drinks:
            self.backend.cancel_drink(d)

        self.assertEquals(first_drink.volume_ml, first_drink.session.volume_ml)
        session_id = first_drink.session.id

        self.backend.cancel_drink(first_drink)

        with self.assertRaises(models.DrinkingSession.DoesNotExist):
            models.DrinkingSession.objects.get(pk=session_id)

        self.assertEquals(num_sessions - 1, models.DrinkingSession.objects.all().count())

    def test_reassign_drink_with_photo(self):
        keg = self.backend.start_keg(METER_NAME, beverage_name=FAKE_BEER_NAME,
                beverage_type='beer', producer_name=FAKE_BREWER_NAME,
                style_name=FAKE_BEER_STYLE)
        self.assertIsNotNone(keg)
        self.assertEquals(0, keg.served_volume())

        drink = self.backend.record_drink(METER_NAME, ticks=1, volume_ml=100,
            photo='foo')

        self.assertTrue(drink.is_guest_pour())
        self.assertTrue(drink.user.is_guest())
        self.assertIsNotNone(drink.picture)
        self.assertTrue(drink.picture.user.is_guest())

        user = self.backend.create_new_user('blort', email='blort@example.com')
        updated_drink = self.backend.assign_drink(drink, user)
        self.assertEquals(user, updated_drink.user)
        self.assertEquals(user, updated_drink.picture.user)

    def test_keg_management(self):
        """Tests adding and removing kegs."""
        tap = models.KegTap.get_from_meter_name(METER_NAME)
        self.assertFalse(tap.is_active(), "Tap is already active.")

        # No beer types yet.
        qs = models.Beverage.objects.filter(name=FAKE_BEER_NAME)
        self.assertEquals(len(qs), 0, "Beverage already exists")

        # Tap the keg.
        keg = self.backend.start_keg(METER_NAME, beverage_name=FAKE_BEER_NAME,
                beverage_type='beer', producer_name=FAKE_BREWER_NAME,
                style_name=FAKE_BEER_STYLE)
        self.assertIsNotNone(keg)
        self.assertTrue(keg.online)

        self.assertEquals(keg.current_tap, tap, "Tap did not become active.")
        tap = keg.current_tap
        self.assertTrue(keg.current_tap.is_active())

        # Check that the beer type was created.
        qs = models.Beverage.objects.filter(name=FAKE_BEER_NAME)
        self.assertEquals(len(qs), 1, "Expected a single new Beverage.")
        beverage = qs[0]
        brewer = beverage.producer
        style = beverage.style
        self.assertEquals(brewer.name, FAKE_BREWER_NAME)
        self.assertEquals(style, FAKE_BEER_STYLE)

        # Now activate a new keg.
        keg = self.backend.end_keg(tap)
        tap = models.KegTap.get_from_meter_name(METER_NAME)
        self.assertFalse(tap.is_active())
        self.assertFalse(keg.online)

        new_keg = self.backend.start_keg(METER_NAME, beverage=beverage)
        tap = models.KegTap.get_from_meter_name(METER_NAME)
        self.assertIsNotNone(new_keg)
        self.assertNotEquals(new_keg, keg)
        self.assertTrue(tap.is_active())
        self.assertTrue(new_keg.online)

        # Ensure the beer type was reused.
        self.assertEquals(new_keg.type, beverage)

        # Deactivate, and activate a new keg again by name.
        self.backend.end_keg(tap)
        new_keg_2 = self.backend.start_keg(METER_NAME, beverage_name='Other Beer',
                beverage_type='beer', producer_name=FAKE_BREWER_NAME,
                style_name=FAKE_BEER_STYLE)
        self.assertEquals(new_keg_2.type.producer, keg.type.producer)
        self.assertEquals(new_keg_2.type.style, keg.type.style)
        self.assertNotEquals(new_keg_2.type, keg.type)

        # New brewer, identical beer name == new beer type.
        tap = models.KegTap.get_from_meter_name(METER_NAME)
        self.backend.end_keg(tap)
        new_keg_3 = self.backend.start_keg(METER_NAME, beverage_name=FAKE_BEER_NAME,
                beverage_type='beer', producer_name='Other Brewer', style_name=FAKE_BEER_STYLE)
        self.assertNotEquals(new_keg_3.type.producer, keg.type.producer)
        self.assertEquals(new_keg_3.type.name, keg.type.name)
        self.assertNotEquals(new_keg_3.type, keg.type)

    def test_meters(self):
        tap = models.KegTap.objects.all()[0]
        meter = tap.current_meter()

        self.assertIsNotNone(meter)
        tap = self.backend.connect_meter(tap, None)
        self.assertIsNone(tap.current_meter())

        tap = self.backend.connect_meter(tap, meter)
        self.assertIsNotNone(tap.current_meter())

    def test_toggles(self):
        tap = models.KegTap.objects.all()[0]
        toggle = tap.current_toggle()

        self.assertIsNotNone(toggle)
        tap = self.backend.connect_toggle(tap, None)
        self.assertIsNone(tap.current_toggle())

        tap = self.backend.connect_toggle(tap, toggle)
        self.assertIsNotNone(tap.current_toggle())

    @override_settings(KEGBOT_BASE_URL='http://example.com:8000//')
    def test_urls(self):
        self.assertEquals('http://example.com:8000', self.backend.get_base_url())

        keg = self.backend.start_keg(METER_NAME, beverage_name=FAKE_BEER_NAME,
                beverage_type='beer', producer_name=FAKE_BREWER_NAME,
                style_name=FAKE_BEER_STYLE)
        self.assertEquals('http://example.com:8000/kegs/{}'.format(keg.id), keg.full_url())

        drink = self.backend.record_drink(METER_NAME, ticks=1, volume_ml=100,
            photo='foo')
        self.assertEquals('http://example.com:8000/d/{}'.format(drink.id), drink.short_url())
        self.assertEquals('http://example.com:8000/s/{}'.format(drink.session.id),
            drink.session.short_url())

        start = drink.session.start_time
        datepart = '{}/{}/{}'.format(start.year, start.month, start.day)
        self.assertEquals('http://example.com:8000/sessions/{}/{}'.format(datepart, drink.session.id),
            drink.session.full_url())



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Django admin site settings for core models."""

from django.contrib import admin
from kegbot.util import util
from pykeg.core import models


class UserAdmin(admin.ModelAdmin):
    list_display = ('username', 'email', 'date_joined', 'last_login', 'is_active',
        'is_superuser', 'is_staff')
    list_filter = ('is_active', 'is_superuser', 'is_staff')
admin.site.register(models.User, UserAdmin)


class KegbotSiteAdmin(admin.ModelAdmin):
    list_display = ('name',)
admin.site.register(models.KegbotSite, KegbotSiteAdmin)


class KegTapAdmin(admin.ModelAdmin):
    list_display = ('name', 'current_keg', 'sort_order')
admin.site.register(models.KegTap, KegTapAdmin)


class KegAdmin(admin.ModelAdmin):
    list_display = ('id', 'type', 'online')
    list_filter = ('online', )
    search_fields = ('id', 'type__name')
admin.site.register(models.Keg, KegAdmin)


class DrinkAdmin(admin.ModelAdmin):
    list_display = ('id', 'user', 'keg', 'time')
    list_filter = ('keg', 'time')
    search_fields = ('id', 'user__username')
admin.site.register(models.Drink, DrinkAdmin)


class AuthenticationTokenAdmin(admin.ModelAdmin):
    list_display = ('auth_device', 'user', 'token_value', 'nice_name', 'enabled', 'IsActive')
    list_filter = ('auth_device', 'enabled')
    search_fields = ('user__username', 'token_value', 'nice_name')
admin.site.register(models.AuthenticationToken, AuthenticationTokenAdmin)


class DrinkingSessionAdmin(admin.ModelAdmin):
    list_display = ('id', 'start_time', 'end_time', 'volume_ml', 'GetTitle')
    list_filter = ('start_time',)
    search_fields = ('name',)
admin.site.register(models.DrinkingSession, DrinkingSessionAdmin)


class ThermoSensorAdmin(admin.ModelAdmin):
    list_display = ('raw_name', 'nice_name')
    search_fields = list_display
admin.site.register(models.ThermoSensor, ThermoSensorAdmin)


def thermolog_deg_c(obj):
    return '%.2f C' % (obj.temp,)


def thermolog_deg_f(obj):
    return '%.2f F' % (util.CtoF(obj.temp),)


class ThermologAdmin(admin.ModelAdmin):
    list_display = ('sensor', thermolog_deg_c, thermolog_deg_f, 'time')
    list_filter = ('sensor', 'time')

admin.site.register(models.Thermolog, ThermologAdmin)


class SystemEventAdmin(admin.ModelAdmin):
    list_display = ('id', 'kind', 'time', 'user', 'drink', 'keg', 'session')
    list_filter = ('kind', 'time')
admin.site.register(models.SystemEvent, SystemEventAdmin)


class PictureAdmin(admin.ModelAdmin):
    list_display = ('id', 'time', 'user', 'keg', 'session', 'caption')
    list_filter = ('time',)
admin.site.register(models.Picture, PictureAdmin)

admin.site.register(models.Beverage)
admin.site.register(models.BeverageProducer)

admin.site.register(models.Controller)
admin.site.register(models.FlowMeter)
admin.site.register(models.FlowToggle)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.core.mail import EmailMultiAlternatives
from django.template import Context
from django.template.loader import get_template

import logging

logger = logging.getLogger('util')

SEPARATOR = '---EMAIL-BLOCK---'


def build_message(to_address, template_name, context_dict):
    from_address = getattr(settings, 'EMAIL_FROM_ADDRESS', None)
    if not from_address:
        logger.error('EMAIL_FROM_ADDRESS is not available; aborting!')
        return None

    template = get_template(template_name)
    context = Context(context_dict)
    rendered = template.render(context)

    parts = (x.strip() for x in rendered.split(SEPARATOR))
    subject, body_plain, body_html, footer_plain, footer_html = parts
    body_plain += '\n\n' + footer_plain
    body_html += '\n\n' + footer_html

    message = EmailMultiAlternatives(subject, body_plain, from_address, [to_address])
    message.attach_alternative(body_html, "text/html")
    return message



from django.conf.urls import include
from django.conf.urls import patterns
from django.conf.urls import url
from django.contrib.auth import views as auth_views

from pykeg.web.kbregistration.forms import PasswordResetForm

urlpatterns = patterns('pykeg.web.kbregistration.views',
    url(r'^register/?$', 'register', name='registration_register'),
)

urlpatterns += patterns('',
                       url(r'^password/change/$',
                           auth_views.password_change,
                           name='password_change'),
                       url(r'^password/change/done/$',
                           auth_views.password_change_done,
                           name='password_change_done'),
                       url(r'^password/reset/$',
                           auth_views.password_reset,
                           kwargs={'password_reset_form': PasswordResetForm},
                           name='password_reset'),
                       url(r'^password/reset/done/$',
                           auth_views.password_reset_done,
                           name='password_reset_done'),
                       url(r'^password/reset/complete/$',
                           auth_views.password_reset_complete,
                           name='password_reset_complete'),
                       url(r'^password/reset/confirm/(?P<uidb64>[0-9A-Za-z]+)-(?P<token>.+)/$',
                           auth_views.password_reset_confirm,
                           name='password_reset_confirm'),

                       url('', include('registration.auth_urls')))



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding field 'Beverage.color_hex'
        db.add_column(u'core_beverage', 'color_hex',
                      self.gf('django.db.models.fields.CharField')(default='#C35900', max_length=16),
                      keep_default=False)

        # Adding field 'Beverage.srm'
        db.add_column(u'core_beverage', 'srm',
                      self.gf('django.db.models.fields.FloatField')(null=True, blank=True),
                      keep_default=False)

        # Adding field 'Beverage.ibu'
        db.add_column(u'core_beverage', 'ibu',
                      self.gf('django.db.models.fields.FloatField')(null=True, blank=True),
                      keep_default=False)

        # Adding field 'Beverage.star_rating'
        db.add_column(u'core_beverage', 'star_rating',
                      self.gf('django.db.models.fields.FloatField')(null=True, blank=True),
                      keep_default=False)


    def backwards(self, orm):
        # Deleting field 'Beverage.color_hex'
        db.delete_column(u'core_beverage', 'color_hex')

        # Deleting field 'Beverage.srm'
        db.delete_column(u'core_beverage', 'srm')

        # Deleting field 'Beverage.ibu'
        db.delete_column(u'core_beverage', 'ibu')

        # Deleting field 'Beverage.star_rating'
        db.delete_column(u'core_beverage', 'star_rating')


    models = {
        u'core.apikey': {
            'Meta': {'object_name': 'ApiKey'},
            'active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'created_time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'device': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Device']", 'null': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.CharField', [], {'default': "'7997680740795478d88b13b857efea51'", 'unique': 'True', 'max_length': '127'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.User']", 'null': 'True', 'blank': 'True'})
        },
        u'core.authenticationtoken': {
            'Meta': {'unique_together': "(('auth_device', 'token_value'),)", 'object_name': 'AuthenticationToken'},
            'auth_device': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            'created_time': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'enabled': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'expire_time': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'nice_name': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'pin': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'token_value': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'tokens'", 'null': 'True', 'to': u"orm['core.User']"})
        },
        u'core.beverage': {
            'Meta': {'ordering': "('name',)", 'object_name': 'Beverage'},
            'abv_percent': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'beverage_backend': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'beverage_backend_id': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'beverage_type': ('django.db.models.fields.CharField', [], {'default': "'beer'", 'max_length': '32'}),
            'calories_per_ml': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'carbs_per_ml': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'color_hex': ('django.db.models.fields.CharField', [], {'default': "'#C35900'", 'max_length': '16'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'ibu': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'original_gravity': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'picture': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Picture']", 'null': 'True', 'blank': 'True'}),
            'producer': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.BeverageProducer']"}),
            'specific_gravity': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'srm': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'star_rating': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'style': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'untappd_beer_id': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'}),
            'vintage_year': ('django.db.models.fields.DateField', [], {'null': 'True', 'blank': 'True'})
        },
        u'core.beverageproducer': {
            'Meta': {'ordering': "('name',)", 'object_name': 'BeverageProducer'},
            'beverage_backend': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'beverage_backend_id': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'country': ('pykeg.core.fields.CountryField', [], {'default': "'USA'", 'max_length': '3'}),
            'description': ('django.db.models.fields.TextField', [], {'default': "''", 'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_homebrew': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'origin_city': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'origin_state': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'picture': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Picture']", 'null': 'True', 'on_delete': 'models.SET_NULL', 'blank': 'True'}),
            'url': ('django.db.models.fields.URLField', [], {'default': "''", 'max_length': '200', 'null': 'True', 'blank': 'True'})
        },
        u'core.controller': {
            'Meta': {'object_name': 'Controller'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model_name': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '128'}),
            'serial_number': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'})
        },
        u'core.device': {
            'Meta': {'object_name': 'Device'},
            'created_time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'default': "'Unknown Device'", 'max_length': '255'})
        },
        u'core.drink': {
            'Meta': {'ordering': "('-time',)", 'object_name': 'Drink'},
            'duration': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'drinks'", 'on_delete': 'models.PROTECT', 'to': u"orm['core.Keg']"}),
            'picture': ('django.db.models.fields.related.OneToOneField', [], {'to': u"orm['core.Picture']", 'unique': 'True', 'null': 'True', 'on_delete': 'models.SET_NULL', 'blank': 'True'}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'drinks'", 'null': 'True', 'on_delete': 'models.PROTECT', 'to': u"orm['core.DrinkingSession']"}),
            'shout': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'tick_time_series': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'ticks': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'time': ('django.db.models.fields.DateTimeField', [], {}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'drinks'", 'to': u"orm['core.User']"}),
            'volume_ml': ('django.db.models.fields.FloatField', [], {})
        },
        u'core.drinkingsession': {
            'Meta': {'ordering': "('-start_time',)", 'object_name': 'DrinkingSession'},
            'end_time': ('django.db.models.fields.DateTimeField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'start_time': ('django.db.models.fields.DateTimeField', [], {}),
            'volume_ml': ('django.db.models.fields.FloatField', [], {'default': '0'})
        },
        u'core.flowmeter': {
            'Meta': {'unique_together': "(('controller', 'port_name'),)", 'object_name': 'FlowMeter'},
            'controller': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'meters'", 'to': u"orm['core.Controller']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'port_name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'tap': ('django.db.models.fields.related.OneToOneField', [], {'blank': 'True', 'related_name': "'meter'", 'unique': 'True', 'null': 'True', 'to': u"orm['core.KegTap']"}),
            'ticks_per_ml': ('django.db.models.fields.FloatField', [], {'default': '5.4'})
        },
        u'core.flowtoggle': {
            'Meta': {'unique_together': "(('controller', 'port_name'),)", 'object_name': 'FlowToggle'},
            'controller': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'toggles'", 'to': u"orm['core.Controller']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'port_name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'tap': ('django.db.models.fields.related.OneToOneField', [], {'blank': 'True', 'related_name': "'toggle'", 'unique': 'True', 'null': 'True', 'to': u"orm['core.KegTap']"})
        },
        u'core.invitation': {
            'Meta': {'object_name': 'Invitation'},
            'expires_date': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime(2014, 8, 22, 0, 0)'}),
            'for_email': ('django.db.models.fields.EmailField', [], {'max_length': '75'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'invite_code': ('django.db.models.fields.CharField', [], {'default': "'f6f3b957-c5fe-487b-adbd-c67dfcb2910d'", 'unique': 'True', 'max_length': '255'}),
            'invited_by': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.User']", 'null': 'True', 'on_delete': 'models.SET_NULL'}),
            'invited_date': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'})
        },
        u'core.keg': {
            'Meta': {'object_name': 'Keg'},
            'description': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'end_time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'finished': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'full_volume_ml': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg_type': ('django.db.models.fields.CharField', [], {'default': "'half-barrel'", 'max_length': '32'}),
            'notes': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'online': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'served_volume_ml': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            'spilled_ml': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            'start_time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Beverage']", 'on_delete': 'models.PROTECT'})
        },
        u'core.kegbotsite': {
            'Meta': {'object_name': 'KegbotSite'},
            'background_image': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Picture']", 'null': 'True', 'on_delete': 'models.SET_NULL', 'blank': 'True'}),
            'check_for_updates': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'google_analytics_id': ('django.db.models.fields.CharField', [], {'max_length': '64', 'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_setup': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'name': ('django.db.models.fields.CharField', [], {'default': "'default'", 'unique': 'True', 'max_length': '64'}),
            'privacy': ('django.db.models.fields.CharField', [], {'default': "'public'", 'max_length': '63'}),
            'registration_id': ('django.db.models.fields.TextField', [], {'default': "''", 'max_length': '128', 'blank': 'True'}),
            'registration_mode': ('django.db.models.fields.CharField', [], {'default': "'public'", 'max_length': '63'}),
            'server_version': ('django.db.models.fields.CharField', [], {'max_length': '64', 'null': 'True'}),
            'session_timeout_minutes': ('django.db.models.fields.PositiveIntegerField', [], {'default': '180'}),
            'temperature_display_units': ('django.db.models.fields.CharField', [], {'default': "'f'", 'max_length': '64'}),
            'timezone': ('django.db.models.fields.CharField', [], {'default': "'UTC'", 'max_length': '255'}),
            'title': ('django.db.models.fields.CharField', [], {'default': "'My Kegbot'", 'max_length': '64'}),
            'volume_display_units': ('django.db.models.fields.CharField', [], {'default': "'imperial'", 'max_length': '64'})
        },
        u'core.kegtap': {
            'Meta': {'ordering': "('sort_order', 'id')", 'object_name': 'KegTap'},
            'current_keg': ('django.db.models.fields.related.OneToOneField', [], {'related_name': "'current_tap'", 'null': 'True', 'on_delete': 'models.SET_NULL', 'to': u"orm['core.Keg']", 'blank': 'True', 'unique': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'sort_order': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'temperature_sensor': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.ThermoSensor']", 'null': 'True', 'on_delete': 'models.SET_NULL', 'blank': 'True'})
        },
        u'core.notificationsettings': {
            'Meta': {'unique_together': "(('user', 'backend'),)", 'object_name': 'NotificationSettings'},
            'backend': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg_ended': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'keg_tapped': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'keg_volume_low': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'session_started': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.User']"})
        },
        u'core.picture': {
            'Meta': {'object_name': 'Picture'},
            'caption': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'image': ('django.db.models.fields.files.ImageField', [], {'max_length': '100'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'pictures'", 'null': 'True', 'on_delete': 'models.SET_NULL', 'to': u"orm['core.Keg']"}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'pictures'", 'null': 'True', 'on_delete': 'models.SET_NULL', 'to': u"orm['core.DrinkingSession']"}),
            'time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'pictures'", 'null': 'True', 'to': u"orm['core.User']"})
        },
        u'core.plugindata': {
            'Meta': {'unique_together': "(('plugin_name', 'key'),)", 'object_name': 'PluginData'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.CharField', [], {'max_length': '127'}),
            'plugin_name': ('django.db.models.fields.CharField', [], {'max_length': '127'}),
            'value': ('pykeg.core.jsonfield.JSONField', [], {})
        },
        u'core.stats': {
            'Meta': {'unique_together': "(('drink', 'user', 'keg', 'session'),)", 'object_name': 'Stats'},
            'drink': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Drink']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_first': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'stats'", 'null': 'True', 'to': u"orm['core.Keg']"}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'stats'", 'null': 'True', 'to': u"orm['core.DrinkingSession']"}),
            'stats': ('pykeg.core.jsonfield.JSONField', [], {}),
            'time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'stats'", 'null': 'True', 'to': u"orm['core.User']"})
        },
        u'core.systemevent': {
            'Meta': {'ordering': "('-id',)", 'object_name': 'SystemEvent'},
            'drink': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'events'", 'null': 'True', 'to': u"orm['core.Drink']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'events'", 'null': 'True', 'to': u"orm['core.Keg']"}),
            'kind': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'events'", 'null': 'True', 'to': u"orm['core.DrinkingSession']"}),
            'time': ('django.db.models.fields.DateTimeField', [], {}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'events'", 'null': 'True', 'to': u"orm['core.User']"})
        },
        u'core.thermolog': {
            'Meta': {'ordering': "('-time',)", 'object_name': 'Thermolog'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'sensor': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.ThermoSensor']"}),
            'temp': ('django.db.models.fields.FloatField', [], {}),
            'time': ('django.db.models.fields.DateTimeField', [], {})
        },
        u'core.thermosensor': {
            'Meta': {'object_name': 'ThermoSensor'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'nice_name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'raw_name': ('django.db.models.fields.CharField', [], {'max_length': '256'})
        },
        u'core.user': {
            'Meta': {'object_name': 'User'},
            'activation_key': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'display_name': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '127'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'mugshot': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'user_mugshot'", 'null': 'True', 'on_delete': 'models.SET_NULL', 'to': u"orm['core.Picture']"}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        }
    }

    complete_apps = ['core']


# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""High-level interface to Django backend."""

from __future__ import absolute_import

import datetime
import logging

from django.conf import settings
from django.db import transaction
from django.utils import timezone
from pykeg.core import keg_sizes
from pykeg.core.util import SuppressTaskErrors
from pykeg.web.auth import get_auth_backend
from pykeg.web.auth import AuthException
from pykeg.web.auth import UserExistsException

from pykeg.core import kb_common
from pykeg.core import models
from pykeg.core import time_series
from pykeg.core.util import get_current_request

from pykeg.backend import exceptions
from pykeg.web import tasks

from . import signals


class UnknownBaseUrlException(Exception):
    pass


class KegbotBackend(object):
    """Provides high-level operations against the Kegbot system."""

    def __init__(self):
        self._logger = logging.getLogger('backend')

    def get_base_url(self):
        """Returns the base URL of the site without a trailing slash.

        Result is used primarily in constructing absolute links back to the
        site, eg in notifications.
        """
        static_url = getattr(settings, 'KEGBOT_BASE_URL', None)
        if static_url:
            return static_url.rstrip('/')
        r = get_current_request()
        if not r:
            raise UnknownBaseUrlException('Cannot determine current request')
        return r.build_absolute_uri('/').rstrip('/')

    @transaction.atomic
    def create_new_user(self, username, email, password=None, photo=None):
        """Creates and returns a User for the given username."""
        try:
            user = get_auth_backend().register(username=username, email=email,
                    password=password, photo=photo)
            signals.user_created.send_robust(sender=self.__class__, user=user)
            return user
        except UserExistsException as e:
            raise exceptions.UserExistsError(e)
        except AuthException as e:
            raise exceptions.BackendError(e)

    @transaction.atomic
    def create_tap(self, name, meter_name=None, toggle_name=None, ticks_per_ml=None):
        """Creates and returns a new KegTap.

        Args:
          name: The human-meaningful name for the tap, for instance, "Main Tap".
          meter_name: The unique sensor name for the tap, for instance,
              "kegboard.flow0".
          toggle_name: If the tap is connected to a relay, this specifies its
              name, for instance, "kegboard.relay0".
          ticks_per_ml: The number of flow meter ticks per milliliter of fluid
              on this tap's meter.

        Returns:
            The new KegTap instance.
        """
        tap = models.KegTap.objects.create(name=name)
        tap.save()

        if meter_name:
            meter = models.FlowMeter.get_or_create_from_meter_name(meter_name)
            meter.tap = tap
            if ticks_per_ml:
                meter.ticks_per_ml = ticks_per_ml
            meter.save()

        if toggle_name:
            toggle = models.FlowToggle.get_or_create_from_toggle_name(toggle_name)
            self.connect_toggle(tap, toggle)

        signals.tap_created.send_robust(sender=self.__class__, tap=tap)
        return tap

    @transaction.atomic
    def create_auth_token(self, auth_device, token_value, username=None):
        """Creates a new AuthenticationToken.

        The combination of (auth_device, token_value) must be unique within the
        system.

        Args:
            auth_device: The name of the authentication device, for instance,
                "core.rfid".
            token_value: The opaque string value of the token, which is typically
                unique within the `auth_device` namespace.
            username: The User with which to associate this Token.

        Returns:
            The newly-created AuthenticationToken.
        """
        token = models.AuthenticationToken.objects.create(
            auth_device=auth_device, token_value=token_value)
        if username:
            user = self.get_user(username)
            token.user = user
        token.save()
        signals.auth_token_created.send_robust(sender=self.__class__, token=token)
        return token

    @transaction.atomic
    def record_drink(self, tap, ticks, volume_ml=None, username=None,
            pour_time=None, duration=0, shout='', tick_time_series='', photo=None,
            spilled=False):
        """Records a new drink against a given tap.

        The tap must have a Keg assigned to it (KegTap.current_keg), and the keg
        must be active.

        Args:
            tap: A KegTap or matching meter name.
            ticks: The number of ticks observed by the flow sensor for this drink.
            volume_ml: The volume, in milliliters, of the pour.  If specifed, this
                value is saved as the Drink's actual value.  If not specified, a
                volume is computed based on `ticks` and the meter's
                `ticks_per_ml` setting.
            username: The username with which to associate this Drink, or None for
                an anonymous Drink.
            pour_time: The datetime of the drink.  If not supplied, the current
                date and time will be used.
            duration: Number of seconds it took to pour this Drink.  This is
                optional information not critical to the record.
            shout: A short comment left by the user about this Drink.  Optional.
            tick_time_series: Vector of flow update data, used for diagnostic
                purposes.
            spilled: If drink is recorded as spill, the volume is added to spilled_ml
                and the "drink" is not saved as an event.

        Returns:
            The newly-created Drink instance.
        """
        tap = self._get_tap(tap)
        if not tap.is_active or not tap.current_keg:
            raise exceptions.BackendError("No active keg at this tap")

        keg = tap.current_keg

        if spilled:
            keg.spilled_ml += volume_ml
            keg.save(update_fields=['spilled_ml'])
            return None

        if volume_ml is None:
            meter = tap.current_meter()
            if not meter:
                raise exceptions.BackendError("Tap has no meter, can't compute volume")
            volume_ml = float(ticks) / meter.ticks_per_ml

        user = None
        if username:
            user = self.get_user(username)
        else:
            user = models.User.objects.get(username='guest')

        if not pour_time:
            pour_time = timezone.now()

        if tick_time_series:
            try:
                # Validate the time series by parsing it; canonicalize it by generating
                # it again.  If malformed, just junk it; it's non-essential information.
                tick_time_series = time_series.to_string(time_series.from_string(tick_time_series))
            except ValueError as e:
                self._logger.warning('Time series invalid, ignoring. Error was: %s' % e)
                tick_time_series = ''

        d = models.Drink(ticks=ticks, keg=keg, user=user,
            volume_ml=volume_ml, time=pour_time, duration=duration,
            shout=shout, tick_time_series=tick_time_series)
        models.DrinkingSession.AssignSessionForDrink(d)
        d.save()

        keg.served_volume_ml += volume_ml
        keg.save(update_fields=['served_volume_ml'])

        if photo:
            pic = models.Picture.objects.create(
                image=photo,
                user=d.user,
                keg=d.keg,
                session=d.session
            )
            d.picture = pic
            d.save()

        events = models.SystemEvent.build_events_for_drink(d)
        tasks.schedule_tasks(events)
        self.build_stats(d.id)

        signals.drink_recorded.send_robust(sender=self.__class__, drink=d)
        return d

    @transaction.atomic
    def cancel_drink(self, drink, spilled=False):
        """Permanently deletes a Drink from the system.

        Associated data, such as SystemEvent, Pictures, and other data, will
        be destroyed along with the drink. Statistics and session data will be
        recomputed after the drink is destroyed.

        Args:
            drink_id: The Drink, or id of the Drink, to cancel.
            spilled: If True, after canceling the Drink, the drink's volume will
                be added to its Keg's "spilled" total.  This is is typically useful
                for removing test pours from the system while still accounting for
                the lost volume.

        Returns:
            The deleted drink.
        """
        if not isinstance(drink, models.Drink):
            drink = models.Drink.objects.get(id=drink)

        session = drink.session
        drink_id = drink.id
        keg = drink.keg
        volume_ml = drink.volume_ml

        keg_update_fields = ['served_volume_ml']
        keg.served_volume_ml -= volume_ml

        # Transfer volume to spillage if requested.
        if spilled and volume_ml and drink.keg:
            keg.spilled_ml += volume_ml
            keg_update_fields.append('spilled_ml')

        keg.save(update_fields=keg_update_fields)

        # Delete the drink, including any objects related to it.
        drink.delete()
        session.Rebuild()

        self.rebuild_stats(drink_id)
        signals.drink_canceled.send_robust(sender=self.__class__, drink=drink)
        return drink

    @transaction.atomic
    def cancel_keg(self, keg):
        """Permanently deletes a keg and ALL drinks."""
        keg_drinks = keg.drinks.all().order_by('id')

        if keg_drinks:
            first_drink_id = keg_drinks[0].id
            sessions = set()
            for drink in keg_drinks:
                sessions.add(drink.session)
                drink.delete()
            for session in sessions:
                session.Rebuild()
            self.rebuild_stats(first_drink_id)
        keg.delete()

    @transaction.atomic
    def assign_drink(self, drink, user):
        """Assigns, or re-assigns, a previously-recorded Drink.

        Statistics and session data will be recomputed as a side-effect
        of any change to user assignment.  (If the drink is already assigned
        to the given user, this method is a no-op).

        Args:
            drink: The Drink object (or drink id) to assign.
            user: The User object (or username) to set as the owner,
                or None for anonymous.

        Returns:
            The drink.
        """
        drink = self.get_drink(drink)
        previous_user = drink.user
        user = self.get_user(user)
        if previous_user == user:
            return drink

        drink.user = user
        drink.save()

        for e in drink.events.all():
            e.user = user
            e.save()
        if drink.picture:
            drink.picture.user = user
            drink.picture.save()

        drink.session.Rebuild()
        self.rebuild_stats(drink.id)

        signals.drink_assigned.send_robust(sender=self.__class__, drink=drink,
            previous_user=previous_user)
        return drink

    def set_drink_volume(self, drink, volume_ml):
        """Updates the drink volume."""
        if volume_ml == drink.volume_ml:
            return

        previous_volume = drink.volume_ml

        with transaction.atomic():
            difference = volume_ml - drink.volume_ml
            drink.volume_ml = volume_ml
            drink.save(update_fields=['volume_ml'])

            keg = drink.keg
            keg.served_volume_ml += difference
            keg.save(update_fields=['served_volume_ml'])

            drink.session.Rebuild()
            self.rebuild_stats(drink.id)

        signals.drink_adjusted.send_robust(sender=self.__class__, drink=drink,
            previous_volume=previous_volume)

    @transaction.atomic
    def log_sensor_reading(self, sensor_name, temperature, when=None):
        """Logs a ThermoSensor reading.

        To avoid an excessive number of entries, the system limits temperature
        readings to one per minute.  If there is already a recording for the
        given time period, that record will be updated with the current temperature
        ("last one wins").

        Regardless of this record's timestamp, any records older than
        `kb_common.THERMO_SENSOR_HISTORY_MINUTES` will be deleted as a side effect
        of this call.

        Args:
            sensor_name: The name of the sensor, corresponding to
                ThermoSensor.raw_name.
            temperature: Temperature, in celsius degrees.  Values outside of the
                range specified in `kb_common.THERMO_SENSOR_RANGE` will be
                rejected.
            when: If specified, a datetime of the recording, otherwise the current
                time is used.

        Returns:
            The record for this reading.
        """
        now = timezone.now()
        if not when:
            when = now

        # The maximum resolution of ThermoSensor records is 1 minute.  Round the
        # time down to the nearest minute; if a record already exists for this time,
        # replace it.
        when = when.replace(second=0, microsecond=0)

        # If the temperature is out of bounds, reject it.
        min_val = kb_common.THERMO_SENSOR_RANGE[0]
        max_val = kb_common.THERMO_SENSOR_RANGE[1]
        if temperature < min_val or temperature > max_val:
            raise ValueError('Temperature out of bounds')

        sensor = self._get_sensor_from_name(sensor_name)
        log_defaults = {
            'temp': temperature,
        }
        record, created = models.Thermolog.objects.get_or_create(sensor=sensor,
          time=when, defaults=log_defaults)
        record.temp = temperature
        record.save()

        # Delete old entries.
        keep_time = now - datetime.timedelta(
            minutes=kb_common.THERMO_SENSOR_HISTORY_MINUTES)
        old_entries = models.Thermolog.objects.filter(time__lt=keep_time)
        old_entries.delete()

        signals.temperature_recorded.send_robust(sender=self.__class__, record=record)
        return record

    @transaction.atomic
    def get_auth_token(self, auth_device, token_value):
        """Fetches the AuthenticationToken matching the given parameters.

        Args:
            auth_device: The requested token's auth_device.
            token_value: The token's value, typically unique within auth_device.

        Returns:
            The matching token.

        Raises:
            NoTokenError: No matching token could be found.
        """
        if token_value and auth_device in kb_common.AUTH_MODULE_NAMES_HEX_VALUES:
            token_value = token_value.lower()
        try:
            return models.AuthenticationToken.objects.get(auth_device=auth_device,
                token_value=token_value)
        except models.AuthenticationToken.DoesNotExist:
            # TODO(mikey): return None instead of raising.
            raise exceptions.NoTokenError

    @transaction.atomic
    def start_keg(self, tap, beverage=None, keg_type=keg_sizes.HALF_BARREL,
            full_volume_ml=None, beverage_name=None, beverage_type=None,
            producer_name=None, style_name=None, when=None):
        """Creates and attaches a new keg."""
        tap = self._get_tap(tap)
        keg = self.create_keg(beverage, keg_type, full_volume_ml, beverage_name, beverage_type,
                producer_name, style_name, notes=None, description=None)
        return self.attach_keg(tap, keg)

    @transaction.atomic
    def attach_keg(self, tap, keg):
        """Activates a keg at the given tap (with existing choices).

        The tap must be inactive (tap.current_keg == None), otherwise a
        ValueError will be thrown.

        Args:
            tap: The KegTap or meter name to tap against.
            keg_id: The Keg ID to map to tap.

        Returns:
            The keg instance.
        """
        tap = self._get_tap(tap)

        if tap.is_active():
            raise ValueError('Tap is already active, must end keg first.')

        keg.start_time = timezone.now()
        keg.online = True
        keg.save()

        old_keg = tap.current_keg
        if old_keg:
            self.end_keg(tap)

        tap.current_keg = keg
        tap.save()

        events = models.SystemEvent.build_events_for_keg(keg)
        tasks.schedule_tasks(events)

        signals.keg_attached.send_robust(sender=self.__class__, keg=keg, tap=tap)
        return keg

    @transaction.atomic
    def end_keg(self, tap):
        """Takes the current Keg offline at the given tap.

        Args:
            tap: The KegTap object to tap against, or a string matching
                the tap's meter name.

        Returns:
            The old keg.

        Raises:
            ValueError: if the tap is missing or already offline.
        """
        tap = self._get_tap(tap)

        if not tap.current_keg:
            raise ValueError('Tap is already offline.')

        keg = tap.current_keg
        keg.online = False
        keg.end_time = timezone.now()
        keg.finished = True
        keg.save()

        tap.current_keg = None
        tap.save()

        events = models.SystemEvent.build_events_for_keg(keg)
        tasks.schedule_tasks(events)

        signals.keg_ended.send_robust(sender=self.__class__, keg=keg)
        return keg

    @transaction.atomic
    def create_keg(self, beverage=None, keg_type=keg_sizes.HALF_BARREL,
            full_volume_ml=None, beverage_name=None, beverage_type=None,
            producer_name=None, style_name=None, notes=None, description=None):
        """Adds a new keg to the keg room (queue).

        A beverage must be specified, either by providing an existing
        Beverage instance as `beverage`, or by specifying values for
        `beverage_type`, `beverage_name`, `producer_name`,
        and `style_name`.

        When using the latter form, the system will attempt to match
        the string type parameters against an already-existing Beverage.
        Otherwise, a new Beverage will be created.

        Args:
            beverage: The type of beverage, as a Beverage object.
            keg_type: The type of physical keg, from keg_sizes.
            full_volume_ml: The keg's original unserved volume.  If unspecified,
                will be interpreted from keg_type.  It is an error to omit this
                parameter when keg_type is OTHER.
            beverage_name: The keg's beverage name.  Must be given with
                `producer_name` and `style_name`;
                `beverage` must be None.
            beverage_type: The keg beverage type.
            producer_name: The brewer or producer of this beverage.
            style_name: The style of this beverage.
            notes: Notes about this keg.
            description: The keg description (private)

        Returns:
            The new keg instance.
        """
        if beverage:
            if beverage_type or beverage_name or producer_name or style_name:
                raise ValueError(
                    'Cannot give beverage_type, beverage_name, producer_name, or style_name with beverage')
        else:
            if not beverage_type:
                raise ValueError('Must supply beverage_type when beverage is None')
            if not beverage_name:
                raise ValueError('Must supply beverage_name when beverage is None')
            if not producer_name:
                raise ValueError('Must supply producer_name when beverage is None')
            if not style_name:
                raise ValueError('Must supply style_name when beverage is None')
            producer = models.BeverageProducer.objects.get_or_create(name=producer_name)[0]
            beverage = models.Beverage.objects.get_or_create(name=beverage_name, beverage_type=beverage_type,
                    producer=producer, style=style_name)[0]

        if keg_type not in keg_sizes.DESCRIPTIONS:
            raise ValueError('Unrecognized keg type: %s' % keg_type)
        if full_volume_ml is None:
            full_volume_ml = keg_sizes.VOLUMES_ML[keg_type]
        else:
            full_volume_ml = full_volume_ml

        when = timezone.now()

        keg = models.Keg.objects.create(type=beverage, keg_type=keg_type,
                online=False, finished=False, full_volume_ml=full_volume_ml,
                start_time=when, end_time=when, notes=notes, description=description)
        signals.keg_created.send_robust(sender=self.__class__, keg=keg)
        return keg

    @transaction.atomic
    def connect_meter(self, tap, new_meter):
        tap = self._get_tap(tap)
        old_meter = tap.current_meter()

        if old_meter != new_meter:
            if old_meter:
                old_meter.tap = None
                old_meter.save()
            if new_meter:
                new_meter.tap = tap
                new_meter.save()

        return tap

    @transaction.atomic
    def connect_toggle(self, tap, new_toggle):
        tap = self._get_tap(tap)
        old_toggle = tap.current_toggle()

        if old_toggle != new_toggle:
            if old_toggle:
                old_toggle.tap = None
                old_toggle.save()
            if new_toggle:
                new_toggle.tap = tap
                new_toggle.save()

        return tap

    def build_stats(self, drink_id):
        """Build statistics for the given drink (or drink), without adjusting
        any future statistics.

        This method should only be called for new recordings (ie, when the newest
        drink record is inserted).  Modifications to existing drinks should call
        `rebuild_stats()`.
        """
        assert drink_id is not None, 'No drink id.'
        with SuppressTaskErrors(self._logger):
            tasks.build_stats.delay(drink_id=drink_id, rebuild_following=False)

    def rebuild_stats(self, drink_id):
        """Rebuild statistics for drink `drink_id`, including statistics for
        all drinks following it.

        Compared to `build_stats()`, this method should be called whenever an
        existing drink has been modified (thus potentially affecting the statistics
        for all later pours).
        """
        assert drink_id is not None, 'No drink id.'
        with SuppressTaskErrors(self._logger):
            tasks.build_stats.delay(drink_id=drink_id, rebuild_following=True)

    def build_backup(self):
        """Builds backup file for Kegbot Site."""
        with SuppressTaskErrors(self._logger):
            tasks.build_backup.delay()

    def _get_tap(self, keg_tap_or_meter_name):
        if isinstance(keg_tap_or_meter_name, models.KegTap):
            return keg_tap_or_meter_name
        try:
            return models.KegTap.get_from_meter_name(keg_tap_or_meter_name)
        except models.KegTap.DoesNotExist, e:
            raise exceptions.BackendError('Invalid tap: %s: %s' % (repr(keg_tap_or_meter_name), e))

    def _get_sensor_from_name(self, name, autocreate=True):
        """Returns the TemperatureSensor with raw_name matching name.

        Args:
            name: The sensor's raw_name.
            autocreate: If True, create a new sensor if one does not exist.

        Returns:
            The TemperatureSensor object, or None if none exists and autocreate was
            not True.
        """
        try:
            return models.ThermoSensor.objects.get(raw_name=name)
        except models.ThermoSensor.DoesNotExist:
            if autocreate:
                sensor = models.ThermoSensor(raw_name=name, nice_name=name)
                sensor.save()
                return sensor
            else:
                return None

    def get_user(self, user_or_username):
        """Returns the User object for the given username, or None."""
        if not user_or_username:
            return None
        if not isinstance(user_or_username, models.User):
            return models.User.objects.get(username=user_or_username)
        return user_or_username

    def get_drink(self, drink_or_id):
        if not isinstance(drink_or_id, models.Drink):
            return models.Drink.objects.get(pk=drink_or_id)
        return drink_or_id



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Generic unittests."""

import os
import subprocess

from django.test import TestCase
from django.utils.importlib import import_module


def path_for_import(name):
    """
    Returns the directory path for the given package or module.
    """
    return os.path.dirname(os.path.abspath(import_module(name).__file__))


class CoreTests(TestCase):

    def test_flake8(self):
        root_path = path_for_import('pykeg')
        command = 'flake8 {}'.format(root_path)
        try:
            subprocess.check_output(command.split())
        except subprocess.CalledProcessError as e:
            print 'command: {}'.format(command)
            print e.output
            self.fail('flake8 failed with return code {}.'.format(e.returncode))



from __future__ import absolute_import

"""Celery beat scheduler backed by Redis.

The schedule will be saved as a pickled data in the key
'celery:beat:<filename>', where filename is the schedule filename
configured in celery

Prerequisite:
  You are using Redis as your broker (BROKER_URL = 'redis://...')

Usage:
  CELERYBEAT_SCHEDULER = 'module.celery.RedisScheduler'
"""

__author__ = 'opensource@hoho.com'

import logging
import pickle

from celery.beat import PersistentScheduler


class RedisShelve(object):
    """Implements the `shelve` interface by pickling to a redis entry."""

    def __init__(self, connection, key_prefix='celery:beat:'):
        self.logger = logging.getLogger(__name__)
        self._connection = connection
        self._key_prefix = key_prefix

        self._filename = None
        self._d = {}

    def open(self, filename, writeback=False, **kwargs):
        assert not self._filename, 'Already open'
        assert not kwargs, 'Unknown arguments: {}'.format(kwargs)

        self._redis_key = '{}{}'.format(self._key_prefix, filename)
        self.logger.debug('open: {}'.format(self._redis_key))

        data = self._connection.get(self._redis_key)
        try:
            data_dict = pickle.loads(data)
        except (EOFError, ValueError, TypeError) as e:
            self.logger.warning('Empty/corrupt persisted data: {}'.format(e))
            data_dict = {}

        class wrapped(dict):
            pass
        wrapped.sync = self.sync
        wrapped.close = self.close

        self._d = wrapped(data_dict)
        return self._d

    def sync(self):
        self.logger.debug('sync: {}'.format(self._d))
        data = pickle.dumps(dict(self._d))
        self._connection.set(self._redis_key, data)

    def close(self):
        self.logger.debug('close')
        self.sync()
        self._d = {}
        self._filename = None


class RedisScheduler(PersistentScheduler):
    """Scheduler that uses a RedisShelve for storage."""

    def __init__(self, *args, **kwargs):
        app = kwargs['app']
        self.persistence = RedisShelve(app.backend.client)
        PersistentScheduler.__init__(self, *args, **kwargs)
